import numpy as np
from astropy.io import fits
from astropy.table import Table
import astropy.units as u
from astropy.io.registry import IORegistryError
from astropy.utils.exceptions import AstropyUserWarning
from astropy.modeling import models
from astropy import coordinates as coord
import gwcs.coordinate_frames as cf
from gwcs.wcs import WCS
import pytest
from asdf import fits_embed

from specutils import Spectrum1D, SpectrumList


# The c1d/x1d reader tests --------------------------

def create_spectrum_hdu(data_len, srctype=None, ver=1, name='EXTRACT1D'):
    """Mock a JWST x1d BinTableHDU"""
    data = np.random.random((data_len, 5))
    table = Table(data=data, names=['WAVELENGTH', 'FLUX', 'ERROR', 'SURF_BRIGHT',
        'SB_ERROR'])

    hdu = fits.BinTableHDU(table, name=name)
    hdu.header['TUNIT1'] = 'um'
    hdu.header['TUNIT2'] = 'Jy'
    hdu.header['TUNIT3'] = 'Jy'
    hdu.header['TUNIT4'] = 'MJy/sr'
    hdu.header['TUNIT5'] = 'MJy/sr'
    hdu.header['SRCTYPE'] = srctype
    hdu.ver = ver

    return hdu


@pytest.fixture(scope="function")
def x1d_single():
    """Mock a JWST x1d HDUList with a single spectrum"""
    hdulist = fits.HDUList()
    hdulist.append(fits.PrimaryHDU())
    hdulist["PRIMARY"].header["TELESCOP"] = ("JWST", "comment")
    # Add a BinTableHDU that contains spectral data
    hdulist.append(create_spectrum_hdu(100, 'POINT', ver=1))
    # Mock the ASDF extension
    hdulist.append(fits.BinTableHDU(name='ASDF'))

    return hdulist


@pytest.fixture(scope="function")
def spec_single(request):
    """Mock a JWST c1d/x1d HDUList with a single spectrum"""
    name = request.param
    hdulist = fits.HDUList()
    hdulist.append(fits.PrimaryHDU())
    hdulist["PRIMARY"].header["TELESCOP"] = ("JWST", "comment")
    # Add a BinTableHDU that contains spectral data
    hdulist.append(create_spectrum_hdu(100, 'POINT', ver=1, name=name))
    # Mock the ASDF extension
    hdulist.append(fits.BinTableHDU(name='ASDF'))

    return hdulist


@pytest.fixture(scope="function")
def spec_multi(request):
    """Mock a JWST c1d/x1d multispec HDUList with 3 spectra"""
    name = request.param
    hdulist = fits.HDUList()
    hdulist.append(fits.PrimaryHDU())
    hdulist["PRIMARY"].header["TELESCOP"] = "JWST"
    # Add a few BinTableHDUs that contain spectral data
    hdulist.append(create_spectrum_hdu(100, 'POINT', ver=1, name=name))
    hdulist.append(create_spectrum_hdu(120, 'EXTENDED', ver=2, name=name))
    hdulist.append(create_spectrum_hdu(110, 'POINT', ver=3, name=name))
    # Mock the ASDF extension
    hdulist.append(fits.BinTableHDU(name='ASDF'))

    return hdulist


@pytest.mark.parametrize('spec_multi, format',
                         [('EXTRACT1D', 'JWST x1d multi'),
                          ('COMBINE1D', 'JWST c1d multi')], indirect=['spec_multi'])
def test_jwst_1d_multi_reader(tmpdir, spec_multi, format):
    """Test SpectrumList.read for JWST c1d/x1d multi data"""
    tmpfile = str(tmpdir.join('jwst.fits'))
    spec_multi.writeto(tmpfile)

    data = SpectrumList.read(tmpfile, format=format)
    assert type(data) is SpectrumList
    assert len(data) == 3

    for item in data:
        assert isinstance(item, Spectrum1D)

    assert data[0].shape == (100,)
    assert data[1].shape == (120,)
    assert data[2].shape == (110,)


@pytest.mark.parametrize('spec_single, format',
                         [('EXTRACT1D', 'JWST x1d'),
                          ('COMBINE1D', 'JWST c1d')], indirect=['spec_single'])
def test_jwst_1d_single_reader(tmpdir, spec_single, format):
    """Test Spectrum1D.read for JWST x1d data"""
    tmpfile = str(tmpdir.join('jwst.fits'))
    spec_single.writeto(tmpfile)

    data = Spectrum1D.read(tmpfile, format=format)
    assert type(data) is Spectrum1D
    assert data.shape == (100,)


@pytest.mark.parametrize("srctype", [None, "UNKNOWN"])
def test_jwst_srctpye_defaults(tmpdir, x1d_single, srctype):
    """ Test """
    tmpfile = str(tmpdir.join('jwst.fits'))

    # Add a spectrum with missing or UNKNOWN SRCTYPE (mutate the fixture)
    x1d_single['EXTRACT1D'].header['SRCTYPE'] == srctype
    x1d_single.writeto(tmpfile)

    data = Spectrum1D.read(tmpfile, format='JWST x1d')
    assert type(data) is Spectrum1D
    assert data.shape == (100,)
    assert x1d_single['EXTRACT1D'].header['SRCTYPE'] == "POINT"


@pytest.mark.parametrize('spec_single', ['EXTRACT1D', 'COMBINE1D'], indirect=['spec_single'])
def test_jwst_1d_single_reader_no_format(tmpdir, spec_single):
    """Test Spectrum1D.read for JWST c1d/x1d data without format arg"""
    tmpfile = str(tmpdir.join('jwst.fits'))
    spec_single.writeto(tmpfile)

    data = Spectrum1D.read(tmpfile)
    assert type(data) is Spectrum1D
    assert data.shape == (100,)
    assert data.unit == u.Jy
    assert data.spectral_axis.unit == u.um


@pytest.mark.parametrize('spec_multi', ['EXTRACT1D', 'COMBINE1D'], indirect=['spec_multi'])
def test_jwst_1d_multi_reader_no_format(tmpdir, spec_multi):
    """Test Spectrum1D.read for JWST c1d/x1d data without format arg"""
    tmpfile = str(tmpdir.join('jwst.fits'))
    spec_multi.writeto(tmpfile)

    data = SpectrumList.read(tmpfile)
    assert type(data) is SpectrumList
    assert len(data) == 3

    for item in data:
        assert isinstance(item, Spectrum1D)


@pytest.mark.parametrize('spec_multi', ['EXTRACT1D', 'COMBINE1D'], indirect=['spec_multi'])
def test_jwst_1d_multi_reader_check_units(tmpdir, spec_multi):
    """Test units for Spectrum1D.read for JWST c1d/x1d data"""
    tmpfile = str(tmpdir.join('jwst.fits'))
    spec_multi.writeto(tmpfile)

    data = SpectrumList.read(tmpfile)
    assert data[0].unit == u.Jy
    assert data[1].unit == u.MJy / u.sr
    assert data[2].unit == u.Jy


@pytest.mark.parametrize('spec_single', ['EXTRACT1D', 'COMBINE1D'], indirect=['spec_single'])
def test_jwst_1d_reader_meta(tmpdir, spec_single):
    """Test that the Primary and COMBINE1D/EXTRACT1D extension headers are merged in meta"""
    tmpfile = str(tmpdir.join('jwst.fits'))
    spec_single.writeto(tmpfile)

    data = Spectrum1D.read(tmpfile)
    assert ('TELESCOP', 'JWST') in data.meta['header'].items()
    assert ('SRCTYPE', 'POINT') in data.meta['header'].items()


@pytest.mark.parametrize('spec_multi', ['EXTRACT1D', 'COMBINE1D'], indirect=['spec_multi'])
def test_jwst_1d_single_reader_fail_on_multi(tmpdir, spec_multi):
    """Make sure Spectrum1D.read on JWST c1d/x1d with many spectra errors out"""
    tmpfile = str(tmpdir.join('jwst.fits'))
    spec_multi.writeto(tmpfile)

    with pytest.raises(IORegistryError):
        Spectrum1D.read(tmpfile)


@pytest.mark.parametrize("srctype", ["BADVAL"])
def test_jwst_reader_fail(tmpdir, x1d_single, srctype):
    """Check that the reader fails when SRCTYPE is a BADVAL"""
    tmpfile = str(tmpdir.join('jwst.fits'))
    hdulist = x1d_single
    # Add a spectrum with bad SRCTYPE (mutate the fixture)
    hdulist.append(create_spectrum_hdu(100, srctype, ver=2))
    hdulist.writeto(tmpfile)

    with pytest.raises(RuntimeError, match="^Keyword"):
        SpectrumList.read(tmpfile, format='JWST x1d multi')


@pytest.mark.xfail(reason="JWST loader no longer attempts to auto-find flux column.")
def test_jwst_reader_warning_stddev(tmpdir, x1d_single):
    """Check that the reader raises warning when stddev is zeros"""
    tmpfile = str(tmpdir.join('jwst.fits'))
    hdulist = x1d_single
    # Put zeros in ERROR column
    hdulist["EXTRACT1D"].data["ERROR"] = 0
    hdulist.writeto(tmpfile)

    with pytest.warns(Warning) as record:
        Spectrum1D.read(tmpfile)
        for r in record:
            if r.message is AstropyUserWarning:
                assert "Standard Deviation has values of 0" in r.message


# The s2d/s3d reader tests -------------------------------

@pytest.fixture
def generate_wcs_transform():
    def _generate_wcs_transform(dispaxis):
        """Create mock gwcs.WCS object for resampled s2d data"""
        detector = cf.Frame2D(name='detector', axes_order=(0, 1), unit=(u.pix, u.pix))
        icrs = cf.CelestialFrame(name='icrs', reference_frame=coord.ICRS(),
            axes_order=(0, 1), unit=(u.deg, u.deg), axes_names=('RA', 'DEC'))
        spec = cf.SpectralFrame(name='spec', axes_order=(2,), unit=(u.micron,),
            axes_names=('lambda',))
        world = cf.CompositeFrame(name="world", frames=[icrs, spec])

        if dispaxis == 1:
            mapping = models.Mapping((0, 1, 0))
        if dispaxis == 2:
            mapping = models.Mapping((0, 1, 1))

        transform = mapping | (models.Const1D(42) & models.Const1D(42)
            & (models.Shift(30) | models.Scale(0.1)))
        pipeline = [(detector, transform),
                    (world, None)]
        wcs = WCS(pipeline)

        return wcs

    return _generate_wcs_transform


@pytest.fixture
def s2d_single(generate_wcs_transform):
    pytest.importorskip("jwst")
    from jwst.datamodels import MultiSlitModel, SlitModel
    from jwst.assign_wcs.util import wcs_bbox_from_shape

    shape = (10, 100)
    dispaxis = 1

    model = MultiSlitModel()
    sm = SlitModel(shape)
    sm.data
    model.slits.append(sm)
    for slit in model.slits:
        slit.meta.wcs = generate_wcs_transform(dispaxis)
        slit.meta.wcs.bounding_box = wcs_bbox_from_shape(shape)
        slit.meta.wcsinfo.dispersion_direction = dispaxis

    model.meta.telescope = "JWST"

    return model


@pytest.fixture(params=[(5, 100), (100, 8)])
def s2d_multi(generate_wcs_transform, request):
    pytest.importorskip("jwst")
    from jwst.datamodels import SlitModel, MultiSlitModel
    from jwst.assign_wcs.util import wcs_bbox_from_shape

    shape = request.param
    if shape[0] < shape[1]:
        dispaxis = 1
    else:
        dispaxis = 2

    model = MultiSlitModel()
    sm = SlitModel(shape)
    sm.data
    model.slits.append(sm)
    model.slits.append(sm)
    for slit in model.slits:
        slit.meta.wcs = generate_wcs_transform(dispaxis)
        slit.meta.wcs.bounding_box = wcs_bbox_from_shape(shape)
        slit.meta.wcsinfo.dispersion_direction = dispaxis
        slit.meta.bunit_data = "Jy"
        slit.meta.bunit_err = "Jy"

    return model


@pytest.mark.xfail(reason="Needs investigation! See #717")
def test_jwst_s2d_reader(tmpdir, s2d_single):
    path = str(tmpdir.join("test.fits"))
    model = s2d_single
    model.save(path)

    spec = Spectrum1D.read(path)
    assert hasattr(spec, "spectral_axis")
    assert spec.unit == u.dimensionless_unscaled


def test_jwst_s2d_multi_reader(tmpdir, s2d_multi):
    path = str(tmpdir.join("test.fits"))
    model = s2d_multi
    model.save(path)

    speclist = SpectrumList.read(path, format="JWST s2d multi")
    assert len(speclist) == 2
    assert hasattr(speclist[0], "spectral_axis")
    assert speclist[1].unit == u.Jy


# The s3d reader tests -------------------------------

def generate_s3d_wcs():
    """ create a fake gwcs for a cube """
    # create input /output frames
    detector = cf.CoordinateFrame(name='detector', axes_order=(0,1,2), axes_names=['x', 'y', 'z'],
                                  axes_type=['spatial', 'spatial', 'spatial'], naxes=3,
                                  unit=['pix', 'pix', 'pix'])
    sky = cf.CelestialFrame(reference_frame=coord.ICRS(), name='sky', axes_names=("RA", "DEC"))
    spec = cf.SpectralFrame(name='spectral', unit=['um'], axes_names=['wavelength'], axes_order=(2,))
    world = cf.CompositeFrame(name="world", frames=[sky, spec])

    # create fake transform to at least get a bounding box
    # for the s3d jwst loader

    # shape 30,10,10 (spec, y, x)
    crpix1, crpix2, crpix3 = 5, 5, 15  # (x, y, spec)
    crval1, crval2, crval3 = 1, 1, 1
    cdelt1, cdelt2, cdelt3 = 0.01, 0.01, 0.05

    shift = models.Shift(-crpix2) & models.Shift(-crpix1)
    scale = models.Multiply(cdelt2) & models.Multiply(cdelt1)
    proj = models.Pix2Sky_TAN()
    skyrot = models.RotateNative2Celestial(crval2, 90 + crval1, 180)
    celestial = shift | scale | proj | skyrot
    wave_model = models.Shift(-crpix3) | models.Multiply(cdelt3) | models.Shift(crval3)
    transform = models.Mapping((2, 0, 1)) | celestial & wave_model | models.Mapping((1, 2, 0))
    # bounding box based on shape (30,10,10) in test
    transform.bounding_box = ((0, 29), (0, 9), (0, 9))

    # create final wcs
    pipeline = [(detector, transform),
                (world, None)]
    return WCS(pipeline)

@pytest.fixture()
def tmp_asdf(tmpdir):
    # Create some data
    sequence = np.arange(100)
    squares  = sequence**2
    random = np.random.random(100)

    # Store the data in an arbitrarily nested dictionary
    tree = {
        'foo': 42,
        'name': 'Monty',
        'sequence': sequence,
        'powers': { 'squares' : squares },
        'random': random,
        'meta': {
            'wcs' : generate_s3d_wcs()
        }
    }

    yield tree
    tree = {}

def create_image_hdu(name='SCI', data=None, shape=None, hdrs=[], ndim=3):
    """ Mock an Image HDU """
    if data is None:
        if not shape:
            shape = [4, 2, 3] if ndim == 3 else [2, 2] if ndim == 2 else [2]
        data = np.zeros(shape)
    hdu = fits.ImageHDU(name=name, data=data, header=fits.Header(hdrs))
    hdu.ver = 1
    return hdu


@pytest.fixture(scope='function')
def cube(tmpdir, tmp_asdf):
    """ Mock a JWST s3d cube """
    hdulist = fits.HDUList()
    hdulist.append(fits.PrimaryHDU())
    hdulist["PRIMARY"].header["TELESCOP"] = ("JWST", "comment")
    hdulist["PRIMARY"].header["FLUXEXT"] = ("ERR", "comment")
    hdulist["PRIMARY"].header["ERREXT"] = ("ERR", "comment")
    hdulist["PRIMARY"].header["MASKEXT"] = ("DQ", "comment")
    # Add ImageHDU for cubes
    shape = [30, 10, 10]
    hdulist.append(create_image_hdu(name='SCI', shape=shape, hdrs=[("BUNIT", 'MJy')]))
    hdulist.append(create_image_hdu(name='ERR', shape=shape,
                                    hdrs=[("BUNIT", 'MJy'), ('ERRTYPE', 'ERR')]))
    hdulist.append(create_image_hdu(name='DQ', shape=shape))

    # Mock the ASDF extension
    hdulist.append(fits.BinTableHDU(name='ASDF'))
    ff = fits_embed.AsdfInFits(hdulist, tmp_asdf)
    tmpfile = str(tmpdir.join('jwst_embedded_asdf.fits'))
    ff.write_to(tmpfile)

    return hdulist


def test_jwst_s3d_single(tmpdir, cube):
    """Test Spectrum1D.read for JWST x1d data"""
    tmpfile = str(tmpdir.join('jwst_s3d.fits'))
    cube.writeto(tmpfile)

    data = Spectrum1D.read(tmpfile, format='JWST s3d')
    assert type(data) is Spectrum1D
    assert data.shape == (10, 10, 30)
    assert data.uncertainty is not None
    assert data.mask is not None
    assert data.uncertainty.unit == 'MJy'



import os

import astropy.units as u
from astropy.nddata import StdDevUncertainty
from astropy.table import Table

from ... import Spectrum1D
from ..registers import data_loader
from ..parsing_utils import (generic_spectrum_from_table,
                             spectrum_from_column_mapping)

__all__ = ['ascii_identify', 'ascii_loader', 'ipac_identify', 'ipac_loader']


def ascii_identify(origin, *args, **kwargs):
    """Check if it's an ASCII file."""
    name = os.path.basename(args[0])

    if name.lower().split('.')[-1] in ['txt', 'ascii']:
        return True

    return False


@data_loader(
    label="ASCII", identifier=ascii_identify, extensions=['txt', 'ascii'],
    priority=-10
)
def ascii_loader(file_name, column_mapping=None, **kwargs):
    """
    Load spectrum from ASCII file.

    Parameters
    ----------
    file_name: str
        The path to the ASCII file.
    column_mapping : dict
        A dictionary describing the relation between the ASCII file columns
        and the arguments of the `Spectrum1D` class, along with unit
        information. The dictionary keys should be the ASCII file column names
        while the values should be a two-tuple where the first element is the
        associated `Spectrum1D` keyword argument, and the second element is the
        unit for the ASCII file column::

            column_mapping = {'FLUX': ('flux', 'Jy')}

    Returns
    -------
    data: Spectrum1D
        The spectrum that is represented by the data in this table.
    """
    tab = Table.read(file_name, format='ascii')

    # If no column mapping is given, attempt to parse the ascii files using
    # unit information
    if column_mapping is None:
        return generic_spectrum_from_table(tab, **kwargs)

    return spectrum_from_column_mapping(tab, column_mapping)


def ipac_identify(*args, **kwargs):
    """Check if it's an IPAC-style ASCII file."""
    name = os.path.basename(args[0])

    if name.lower().split('.')[-1] in ['txt', 'dat']:
        return True

    return False


@data_loader(label="IPAC", identifier=ipac_identify, extensions=['txt', 'dat'])
def ipac_loader(file_name, column_mapping=None, **kwargs):
    """
    Load spectrum from IPAC-style ASCII file

    Parameters
    ----------
    file_name: str
        The path to the IPAC-style ASCII file.
    column_mapping : dict
        A dictionary describing the relation between the IPAC-style ASCII
        file columns and the arguments of the `Spectrum1D` class, along with
        unit information. The dictionary keys should be the IPAC-style ASCII
        file column names while the values should be a two-tuple where the
        first element is the associated `Spectrum1D` keyword argument, and the
        second element is the unit for the IPAC-style ASCII file column::

            column_mapping = {'FLUX': ('flux', 'Jy')}

    Returns
    -------
    data: Spectrum1D
        The spectrum that is represented by the data in this table.
    """
    tab = Table.read(file_name, format='ascii.ipac')

    # If no column mapping is given, attempt to parse the ascii files using
    # unit information
    if column_mapping is None:
        return generic_spectrum_from_table(tab, **kwargs)

    return spectrum_from_column_mapping(tab, column_mapping)



import pickle

import pytest
import numpy as np
from astropy import units as u
from astropy import modeling

from specutils.utils import QuantityModel
from ..utils.wcs_utils import refraction_index, vac_to_air, air_to_vac

wavelengths = [300, 500, 1000] * u.nm
data_index_refraction = {
   'Griesen2006': np.array([3.07393068, 2.9434858 , 2.8925797 ]),
   'Edlen1953': np.array([2.91557413, 2.78963801, 2.74148172]),
   'Edlen1966': np.array([2.91554272, 2.7895973 , 2.74156098]),
   'PeckReeder1972': np.array([2.91554211, 2.78960005, 2.74152561]),
   'Morton2000': np.array([2.91568573, 2.78973402, 2.74169531]),
   'Ciddor1996': np.array([2.91568633, 2.78973811, 2.74166131])
}

def test_quantity_model():
    c = modeling.models.Chebyshev1D(3)
    uc = QuantityModel(c, u.AA, u.km)

    assert uc(10*u.nm).to(u.m) == 0*u.m

def test_pickle_quantity_model(tmp_path):
    """
    Check that a QuantityModel can roundtrip through pickling, as it
    would if fit in a multiprocessing pool.
    """

    c = modeling.models.Chebyshev1D(3)
    uc = QuantityModel(c, u.AA, u.km)

    pkl_file = tmp_path / "qmodel.pkl"

    with open(pkl_file, "wb") as f:
        pickle.dump(uc, f)

    with open(pkl_file, "rb") as f:
        new_model = pickle.load(f)

    assert new_model.input_units == uc.input_units
    assert new_model.return_units == uc.return_units
    assert type(new_model.unitless_model) == type(uc.unitless_model)
    assert np.all(new_model.unitless_model.parameters == uc.unitless_model.parameters)

@pytest.mark.parametrize("method", data_index_refraction.keys())
def test_refraction_index(method):
    tmp = (refraction_index(wavelengths, method) - 1) * 1e4
    assert np.isclose(tmp, data_index_refraction[method], atol=1e-7).all()


@pytest.mark.parametrize("method", data_index_refraction.keys())
def test_air_to_vac(method):
    tmp = refraction_index(wavelengths, method)
    assert np.isclose(wavelengths.value * tmp,
                      air_to_vac(wavelengths, method=method, scheme='inversion').value,
                      rtol=1e-6).all()
    assert np.isclose(wavelengths.value,
                      air_to_vac(vac_to_air(wavelengths, method=method),
                                 method=method, scheme='iteration').value,
                      atol=1e-12).all()



"""
A module for analysis tools focused on determining the location of
spectral features.
"""

import numpy as np

from ..spectra import SpectralRegion
from ..manipulation import extract_region


__all__ = ['centroid']


def centroid(spectrum, region):
    """
    Calculate the centroid of a region, or regions, of the spectrum.

    Parameters
    ----------
    spectrum : `~specutils.spectra.spectrum1d.Spectrum1D`
        The spectrum object overwhich the centroid will be calculated.

    region: `~specutils.utils.SpectralRegion` or list of `~specutils.utils.SpectralRegion`
        Region within the spectrum to calculate the centroid.

    Returns
    -------
    centroid : float or list (based on region input)
        Centroid of the spectrum or within the regions

    Notes
    -----
    The spectrum will need to be continuum subtracted before calling
    this method. See the
    `analysis documentation <https://specutils.readthedocs.io/en/latest/analysis.html>`_ for more information.

    """

    # No region, therefore whole spectrum.
    if region is None:
        return _centroid_single_region(spectrum)

    # Single region
    elif isinstance(region, SpectralRegion):
        return _centroid_single_region(spectrum, region=region)

    # List of regions
    elif isinstance(region, list):
        return [_centroid_single_region(spectrum, region=reg)
                for reg in region]


def _centroid_single_region(spectrum, region=None):
    """
    Calculate the centroid of the spectrum based on the flux and uncertainty
    in the spectrum.

    Parameters
    ----------
    spectrum : `~specutils.spectra.spectrum1d.Spectrum1D`
        The spectrum object overwhich the centroid will be calculated.

    region: `~specutils.utils.SpectralRegion`
        Region within the spectrum to calculate the centroid.

    Returns
    -------
    centroid : float or list (based on region input)
        Centroid of the spectrum or within the regions

    Notes
    -----
    This is a helper function for the above `centroid()` method.

    """

    if region is not None:
        calc_spectrum = extract_region(spectrum, region)
    else:
        calc_spectrum = spectrum

    if hasattr(spectrum, 'mask') and spectrum.mask is not None:
        flux = calc_spectrum.flux[~calc_spectrum.mask]
        dispersion = calc_spectrum.spectral_axis[~calc_spectrum.mask].quantity
    else:
        flux = calc_spectrum.flux
        dispersion = calc_spectrum.spectral_axis.quantity

    if len(flux.shape) > 1:
        dispersion = np.tile(dispersion, [flux.shape[0], 1])

    # the axis=-1 will enable this to run on single-dispersion, single-flux
    # and single-dispersion, multiple-flux
    return np.sum(flux * dispersion, axis=-1) / np.sum(flux, axis=-1)



import astropy.units as u
import numpy as np
import pytest

from ..spectra.spectrum1d import Spectrum1D
from .spectral_examples import simulated_spectra


def test_spectral_axes():
    flux1 = (np.random.sample(49) * 100).astype(int)
    flux2 = (np.random.sample(49) * 100).astype(int)

    flux3 = flux1 + flux2

    spec1 = Spectrum1D(spectral_axis=np.arange(1, 50) * u.nm,
                       flux=flux1 * u.Jy)
    spec2 = Spectrum1D(spectral_axis=np.arange(1, 50) * u.nm,
                       flux=flux2 * u.Jy)

    spec3 = spec1 + spec2

    assert np.allclose(spec3.flux.value, flux3)


def test_add_basic_spectra(simulated_spectra):

    # Get the numpy array of data
    flux1 = simulated_spectra.s1_um_mJy_e1_flux
    flux2 = simulated_spectra.s1_um_mJy_e2_flux
    flux3 = flux1 + flux2

    # Calculate using the spectrum1d/nddata code
    spec3 = simulated_spectra.s1_um_mJy_e1 + simulated_spectra.s1_um_mJy_e2

    assert np.allclose(spec3.flux.value, flux3)


def test_add_diff_flux_prefix(simulated_spectra):

    # Get the numpy array of data
    # this assumes output will be in mJy units
    flux1 = simulated_spectra.s1_AA_mJy_e3_flux
    flux2 = simulated_spectra.s1_AA_nJy_e4_flux
    flux3 = flux1 + (flux2 / 1000000)

    # Calculate using the spectrum1d/nddata code
    spec3 = simulated_spectra.s1_AA_mJy_e3 + simulated_spectra.s1_AA_nJy_e4

    assert np.allclose(spec3.flux.value, flux3)


def test_subtract_basic_spectra(simulated_spectra):

    # Get the numpy array of data
    flux1 = simulated_spectra.s1_um_mJy_e1_flux
    flux2 = simulated_spectra.s1_um_mJy_e2_flux
    flux3 = flux2 - flux1

    # Calculate using the spectrum1d/nddata code
    spec3 = simulated_spectra.s1_um_mJy_e2 - simulated_spectra.s1_um_mJy_e1

    assert np.allclose(spec3.flux.value, flux3)


def test_divide_basic_spectra(simulated_spectra):

    # Get the numpy array of data
    flux1 = simulated_spectra.s1_um_mJy_e1_flux
    flux2 = simulated_spectra.s1_um_mJy_e2_flux
    flux3 = flux1 / flux2

    # Calculate using the spectrum1d/nddata code
    spec3 = simulated_spectra.s1_um_mJy_e1 / simulated_spectra.s1_um_mJy_e2

    assert np.allclose(spec3.flux.value, flux3)


def test_multiplication_basic_spectra(simulated_spectra):

    # Get the numpy array of data
    flux1 = simulated_spectra.s1_um_mJy_e1_flux
    flux2 = simulated_spectra.s1_um_mJy_e2_flux
    flux3 = flux1 * flux2

    # Calculate using the spectrum1d/nddata code
    spec3 = simulated_spectra.s1_um_mJy_e1 * simulated_spectra.s1_um_mJy_e2

    assert np.allclose(spec3.flux.value, flux3)


def test_add_diff_spectral_axis(simulated_spectra):

    # Calculate using the spectrum1d/nddata code
    spec3 = simulated_spectra.s1_um_mJy_e1 + simulated_spectra.s1_AA_mJy_e3


def test_masks(simulated_spectra):
    masked_spec = simulated_spectra.s1_um_mJy_e1_masked

    masked_sum = masked_spec + masked_spec
    assert np.all(masked_sum.mask == simulated_spectra.s1_um_mJy_e1_masked.mask)

    masked_sum.mask[:50] = True
    masked_diff = masked_sum - masked_spec
    assert u.allclose(masked_diff.flux, masked_spec.flux)
    assert np.all(masked_diff.mask == masked_sum.mask | masked_spec.mask)


def test_mask_nans():
    flux1 = np.random.random(10)
    flux2 = np.random.random(10)
    nan_idx = [1, 3, 5]
    flux2[nan_idx] = np.nan
    spec1 = Spectrum1D(spectral_axis=np.arange(10) * u.nm, flux=flux1 * u.Jy)
    spec2 = Spectrum1D(spectral_axis=np.arange(10) * u.nm, flux=flux2 * u.Jy)

    spec3 = spec1 + spec2

    assert spec3.mask[nan_idx].all() == True



from specutils import SpectrumList
from specutils.io.registers import data_loader

from .dc_common import (
    FITS_FILE_EXTS, SINGLE_SPLIT_LABEL, MULTILINE_SINGLE_LABEL,
)
from ..parsing_utils import read_fileobj_or_hdulist

GAMA_2QZ_CONFIG = {
    "hdus": None,
    "wcs": {
        "pixel_reference_point_keyword": "CRPIX1",
        "pixel_reference_point_value_keyword": "CRVAL1",
        "pixel_width_keyword": "CD1_1",
        "wavelength_unit": "Angstrom",
    },
    "units": {"flux_unit": "count"},
    "all_standard_units": False,
    "all_keywords": True,
    "valid_wcs": False,
}
GAMA_2SLAQ_QSO_CONFIG = {
    "hdus": None,
    "wcs": {
        "pixel_reference_point_keyword": "CRPIX1",
        "pixel_reference_point_value_keyword": "CRVAL1",
        "pixel_width_keyword": "CDELT1",
        "wavelength_unit": "Angstrom",
    },
    "units": {"flux_unit": "count"},
    "all_standard_units": False,
    "all_keywords": True,
    "valid_wcs": False,
}
GAMA_CONFIG = {
    "hdu": {
        "1": {"purpose": "science", "units": {
            "flux_unit": "10^-17 erg/s/cm^2/A"
        }, },
        "2": {"purpose": "error_stdev"},
        "3": {"purpose": "unreduced_science"},
        "4": {"purpose": "unreduced_error_stdev"},
        "5": {"purpose": "sky"},
    },
    "wcs": None,
    "units": {"flux_unit": "count"},
    "all_standard_units": False,
    "all_keywords": False,
    "valid_wcs": True,
}
GAMA_LT_CONFIG = {
    "hdus": {"0": {"purpose": "science"}, },
    "wcs": {
        "pixel_reference_point_keyword": "CRPIX",
        "pixel_reference_point_value_keyword": "CRVAL",
        "pixel_width_keyword": "CDELT",
        "wavelength_unit": "Angstrom",
    },
    "units": {"flux_unit": "count"},
    "all_standard_units": False,
    "all_keywords": False,
    "valid_wcs": False,
}
MGC_CONFIG = {
    "hdu": None,
    "units": {"flux_unit": "count"},
    "wcs": None,
    "all_standard_units": False,
    "all_keywords": True,
    "valid_wcs": True,
}
WIGGLEZ_CONFIG = {
    "hdus": {
        "0": {"purpose": "science"},
        "1": {"purpose": "error_variance"},
        "2": {"purpose": "skip"},
    },
    "wcs": {
        "pixel_reference_point_keyword": "CRPIX1",
        "pixel_reference_point_value_keyword": "CRVAL1",
        "pixel_width_keyword": "CDELT1",
        "wavelength_unit": "Angstrom",
    },
    "units": {"flux_unit": "count"},
    "all_standard_units": False,
    "all_keywords": False,
    "valid_wcs": False,
}
GAMA_2DFGRS_CONFIG = {
    "hdu": None,
    "wcs": {
        "pixel_reference_point_keyword": "CRPIX1",
        "pixel_reference_point_value_keyword": "CRVAL1",
        "pixel_width_keyword": "CDELT1",
        "wavelength_unit": "Angstrom",
    },
    "units": {"flux_unit": "count"},
    "all_standard_units": False,
    "all_keywords": True,
    "valid_wcs": False,
}


def identify_2qz(origin, *args, **kwargs):
    """
    Identify if the current file is a OzDES file
    """
    with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
        header = hdulist[0].header
        if "2QZ" in header.get("SURVEY", "") and "GAMANAME" in header:
            return True
        return False


@data_loader(
    label="GAMA-2QZ", extensions=FITS_FILE_EXTS, dtype=SpectrumList,
    identifier=identify_2qz, priority=10,
)
def twoqz_loader(filename):
    spectra = SpectrumList.read(
        filename, format=SINGLE_SPLIT_LABEL, **GAMA_2QZ_CONFIG
    )
    return spectra


def identify_2slaq_qso(origin, *args, **kwargs):
    """
    Identify if the current file is a OzDES file
    """
    with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
        header = hdulist[0].header
        if "2SLAQ-QSO" in header.get("SURVEY", "") and "GAMANAME" in header:
            return True
        return False


@data_loader(
    label="GAMA-2SLAQ-QSO", extensions=FITS_FILE_EXTS, dtype=SpectrumList,
    identifier=identify_2slaq_qso, priority=10,
)
def twoslaq_qso_loader(filename):
    spectra = SpectrumList.read(
        filename, format=SINGLE_SPLIT_LABEL, **GAMA_2SLAQ_QSO_CONFIG
    )
    return spectra


def identify_gama(origin, *args, **kwargs):
    """
    Identify if the current file is a OzDES file
    """
    with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
        header = hdulist[0].header
        if "GAMA" in header.get("ORIGIN", "").strip() and "GAMANAME" in header:
            return True
        return False


@data_loader(
    label="GAMA", extensions=FITS_FILE_EXTS, dtype=SpectrumList,
    identifier=identify_gama, priority=10,
)
def gama_loader(filename):
    spectra = SpectrumList.read(
        filename, format=MULTILINE_SINGLE_LABEL, **GAMA_CONFIG
    )
    return spectra


def identify_gama_lt(origin, *args, **kwargs):
    """
    Identify if the current file is a OzDES file
    """
    with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
        header = hdulist[0].header
        if "Liverpool JMU" in header.get("ORIGIN", "") and "GAMANAME" in header:
            return True
        return False


@data_loader(
    label="GAMA-LT", extensions=FITS_FILE_EXTS, dtype=SpectrumList,
    identifier=identify_gama_lt, priority=10,
)
def gama_lt_loader(filename):
    spectra = SpectrumList.read(
        filename, format=SINGLE_SPLIT_LABEL, **GAMA_LT_CONFIG
    )
    return spectra


def identify_mgc(origin, *args, **kwargs):
    """
    Identify if the current file is a OzDES file
    """
    with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
        header = hdulist[0].header
        if "MGC" in header.get("SURVEY", "") and "GAMANAME" in header:
            return True
        return False


@data_loader(
    label="GAMA-MGC", extensions=FITS_FILE_EXTS, dtype=SpectrumList,
    identifier=identify_mgc, priority=10,
)
def mgc_loader(filename):
    spectra = SpectrumList.read(
        filename, format=MULTILINE_SINGLE_LABEL, **MGC_CONFIG
    )
    return spectra


def identify_wigglez(origin, *args, **kwargs):
    """
    Identify if the current file is a OzDES file
    """
    with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
        header = hdulist[0].header
        if "WiggleZ" in header.get("SURVEY", "") and "GAMANAME" in header:
            return True
        return False


@data_loader(
    label="GAMA-WiggleZ", extensions=FITS_FILE_EXTS, dtype=SpectrumList,
    identifier=identify_wigglez, priority=10,
)
def wigglez_loader(filename):
    spectra = SpectrumList.read(
        filename, format=SINGLE_SPLIT_LABEL, **WIGGLEZ_CONFIG
    )
    return spectra


def identify_2dfgrs(origin, *args, **kwargs):
    """
    Identify if the current file is a OzDES file
    """
    with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
        header = hdulist[0].header
        if "2dFGRS" in header.get("SURVEY", "") and "GAMANAME" in header:
            return True
        return False


@data_loader(
    label="GAMA-2dFGRS", extensions=FITS_FILE_EXTS, dtype=SpectrumList,
    identifier=identify_2dfgrs, priority=10,
)
def gama_2dfgrs_loader(filename):
    spectra = SpectrumList.read(
        filename, format=MULTILINE_SINGLE_LABEL, **GAMA_2DFGRS_CONFIG
    )
    return spectra



import numpy as np
import pytest
import astropy.units as u
from astropy.nddata import InverseVariance, StdDevUncertainty
from astropy.tests.helper import assert_quantity_allclose

from ..spectra.spectrum1d import Spectrum1D
from ..tests.spectral_examples import simulated_spectra
from ..manipulation.resample import FluxConservingResampler, LinearInterpolatedResampler, SplineInterpolatedResampler


@pytest.fixture(params=[FluxConservingResampler, LinearInterpolatedResampler, SplineInterpolatedResampler])
def all_resamplers(request):
    return request.param

# todo: Should add tests for different weighting options once those
# are more solidified.


def test_same_grid_fluxconserving(simulated_spectra):
    """
    Test that feeding in the original dispersion axis returns the
    same flux after resampling.
    """
    input_spectra = simulated_spectra.s1_um_mJy_e1
    input_spectra.uncertainty = InverseVariance([0.5]*len(simulated_spectra.s1_um_mJy_e1.flux))

    inst = FluxConservingResampler()
    results = inst(input_spectra,
                   simulated_spectra.s1_um_mJy_e1.spectral_axis)

    assert np.allclose(np.array(simulated_spectra.s1_um_mJy_e1.flux),
                       np.array(results.flux))
    assert np.allclose(input_spectra.uncertainty.array,
                       results.uncertainty.array)


def test_expanded_grid_fluxconserving():
    """
    New dispersion axis has more bins then input dispersion axis
    """

    flux_val = np.array([1, 3, 7, 6, 20])
    wave_val = np.array([2, 4, 12, 16, 20])
    input_spectra = Spectrum1D(flux=flux_val * u.mJy, spectral_axis=wave_val * u.nm)
    resamp_grid = [1, 5, 9, 13, 14, 17, 21, 22, 23] * u.nm

    inst = FluxConservingResampler()
    results = inst(input_spectra, resamp_grid)

    assert_quantity_allclose(results.flux,
                            np.array([np.nan, 3., 6.13043478, 7., 6.33333333, 10., 20., np.nan, np.nan])*u.mJy)


def test_stddev_uncert_propogation():
    """
    Check uncertainty propagation if input uncertainty is InverseVariance
    """
    flux_val = np.array([1, 3, 7, 6, 20])
    wave_val = np.array([20, 30, 40, 50, 60])
    input_spectra = Spectrum1D(flux=flux_val * u.mJy, spectral_axis=wave_val * u.AA,
                               uncertainty=StdDevUncertainty([0.1, 0.25, 0.1, 0.25, 0.1]))

    inst = FluxConservingResampler()
    results = inst(input_spectra, [25, 35, 50, 55]*u.AA)

    assert np.allclose(results.uncertainty.array,
                       np.array([27.5862069, 38.23529412, 17.46724891, 27.5862069]))


def delta_wl(saxis):
    """
    A helper function that computes the "size" of a bin given the bin centers
    for testing the flux conservation
    """
    l_widths = (saxis[1] - saxis[0])
    r_widths = (saxis[-1] - saxis[-2])
    # if three bins 0,1,2; want width of central bin.  width is avg of 1/2 minus
    # average of 0/1: (i1 + i2)/2 - (i0 + i1)/2 = (i2 - i0)/2
    mid_widths = (saxis[2:] - saxis[:-2]) / 2

    return np.concatenate([[l_widths.value], mid_widths.value, [r_widths.value]])*saxis.unit


@pytest.mark.parametrize("specflux,specwavebins,outwavebins", [
    ([1, 3, 2], [4000, 5000, 6000, 7000], np.linspace(4000, 7000, 5)),
    ([1, 3, 2, 1], np.linspace(4000, 7000, 5), [4000, 5000, 6000, 7000])
    ])
def test_flux_conservation(specflux, specwavebins, outwavebins):
    """
    A few simple cases to programatically ensure flux is conserved in the
    resampling algorithm
    """
    specwavebins = specwavebins*u.AA
    outwavebins = outwavebins*u.AA
    specflux = specflux*u.AB

    specwave = (specwavebins[:-1] + specwavebins[1:])/2
    outwave = (outwavebins[:-1] + outwavebins[1:])/2

    in_spec = Spectrum1D(spectral_axis=specwave, flux=specflux)
    out_spec = FluxConservingResampler()(in_spec, outwave)

    in_dwl = delta_wl(in_spec.spectral_axis)
    out_dwl = delta_wl(out_spec.spectral_axis)

    flux_in = np.sum(in_spec.flux * in_dwl)
    flux_out = np.sum(out_spec.flux * out_dwl)

    assert_quantity_allclose(flux_in, flux_out)


def test_multi_dim_spectrum1D():
    """
    Test for input spectrum1Ds that have a two dimensional flux and
    uncertainty.
    """
    flux_2d = np.array([np.ones(10) * 5, np.ones(10) * 6, np.ones(10) * 7])

    input_spectra = Spectrum1D(spectral_axis=np.arange(5000, 5010) * u.AA,
                      flux=flux_2d * u.Jy,
                      uncertainty=StdDevUncertainty(flux_2d / 10))

    inst = FluxConservingResampler()
    results = inst(input_spectra, [5001, 5003, 5005, 5007] * u.AA)

    assert_quantity_allclose(results.flux,
                            np.array([[5., 5., 5., 5.],
                                      [6., 6., 6., 6.],
                                      [7., 7., 7., 7.]]) * u.Jy)
    assert np.allclose(results.uncertainty.array,
                       np.array([[4., 4., 4., 4.],
                                 [2.77777778, 2.77777778, 2.77777778, 2.77777778],
                                 [2.04081633, 2.04081633, 2.04081633, 2.04081633]] ))


def test_expanded_grid_interp_linear():
    """
    New dispersion axis has more bins then input dispersion axis
    """

    flux_val = np.array([1, 3, 7, 6, 20])
    wave_val = np.array([2, 4, 12, 16, 20])
    input_spectra = Spectrum1D(spectral_axis=wave_val * u.AA, flux=flux_val * u.mJy)
    resamp_grid = [1, 5, 9, 13, 14, 17, 21, 22, 23] * u.AA

    inst = LinearInterpolatedResampler()
    results = inst(input_spectra, resamp_grid)

    assert_quantity_allclose(results.flux,
                            np.array([np.nan, 3.5, 5.5, 6.75, 6.5, 9.5, np.nan, np.nan, np.nan])*u.mJy)


def test_expanded_grid_interp_spline():
    """
    New dispersion axis has more bins then input dispersion axis
    """

    flux_val = np.array([1, 3, 7, 6, 20])
    wave_val = np.array([2, 4, 12, 16, 20])
    input_spectra = Spectrum1D(spectral_axis=wave_val * u.AA, flux=flux_val * u.mJy)
    resamp_grid = [1, 5, 9, 13, 14, 17, 21, 22, 23] * u.AA

    inst = SplineInterpolatedResampler()
    results = inst(input_spectra, resamp_grid)

    assert_quantity_allclose(results.flux,
                            np.array([np.nan, 3.98808594, 6.94042969, 6.45869141,
                                      5.89921875, 7.29736328, np.nan, np.nan, np.nan])*u.mJy)


@pytest.mark.parametrize("edgetype,lastvalue",
                         [("nan_fill", np.nan), ("zero_fill", 0)])
def test_resample_edges(edgetype, lastvalue, all_resamplers):
    input_spectrum = Spectrum1D(spectral_axis=[2, 4, 12, 16, 20] * u.micron,
                                flux=[1, 3, 7, 6, 20] * u.mJy)
    resamp_grid = [1, 3, 7, 6, 20, 100] * u.micron

    resampler = all_resamplers(edgetype)
    resampled = resampler(input_spectrum, resamp_grid)
    if lastvalue is np.nan:
        assert np.isnan(resampled.flux[-1])
    else:
        assert resampled.flux[-1] == lastvalue


def test_resample_different_units(all_resamplers):
    input_spectrum = Spectrum1D(spectral_axis=[5000, 6000 ,7000] * u.AA,
                                flux=[1, 2, 3] * u.mJy)
    resampler = all_resamplers("nan_fill")
    if all_resamplers == FluxConservingResampler:
        pytest.xfail('flux conserving resampler cannot yet handle differing units')


    resamp_grid = [5500, 6500]*u.nm
    resampled = resampler(input_spectrum, resamp_grid)
    assert np.all(np.isnan(resampled.flux))

    resamp_grid = [550, 650]*u.nm
    resampled = resampler(input_spectrum, resamp_grid)
    assert not np.any(np.isnan(resampled.flux))


def test_resample_uncs(all_resamplers):
    sdunc = StdDevUncertainty([0.1,0.2, 0.3]*u.mJy)
    input_spectrum = Spectrum1D(spectral_axis=[5000, 6000 ,7000] * u.AA,
                                flux=[1, 2, 3] * u.mJy,
                                uncertainty=sdunc)

    resampled = all_resamplers()(input_spectrum, [5500, 6500]*u.AA)
    if all_resamplers == FluxConservingResampler:
        # special-cased because it switches the unc to inverse variance by construction
        assert resampled.uncertainty.unit == sdunc.unit**-2
    else:
        assert resampled.uncertainty.unit == sdunc.unit
        assert resampled.uncertainty.uncertainty_type == sdunc.uncertainty_type



import pytest
import numpy as np

import astropy.units as u
from astropy.modeling import models
from astropy.nddata import StdDevUncertainty
from astropy.stats.funcs import gaussian_sigma_to_fwhm
from astropy.tests.helper import quantity_allclose
from astropy.utils.exceptions import AstropyUserWarning

from ..spectra import Spectrum1D, SpectralRegion
from ..spectra.spectrum_collection import SpectrumCollection
from ..spectra.spectral_axis import SpectralAxis

from ..analysis import (line_flux, equivalent_width, snr, centroid,
                        gaussian_sigma_width, gaussian_fwhm, fwhm, moment,
                        snr_derived, fwzi, is_continuum_below_threshold)
from ..fitting import find_lines_threshold
from ..manipulation import snr_threshold, FluxConservingResampler
from ..tests.spectral_examples import simulated_spectra


def test_line_flux():

    np.random.seed(42)

    frequencies = np.linspace(100, 1, 10000) * u.GHz
    g = models.Gaussian1D(amplitude=1*u.Jy, mean=10*u.GHz, stddev=1*u.GHz)
    noise = np.random.normal(0., 0.01, frequencies.shape) * u.Jy
    flux = g(frequencies) + noise

    spectrum = Spectrum1D(spectral_axis=frequencies, flux=flux)

    result = line_flux(spectrum)

    assert result.unit.is_equivalent(u.erg / u.cm**2 / u.s)

    # Account for the fact that Astropy uses a different normalization of the
    # Gaussian where the integral is not 1
    expected = np.sqrt(2*np.pi) * u.GHz * u.Jy

    assert quantity_allclose(result, expected, atol=0.01*u.GHz*u.Jy)


def test_line_flux_uncertainty():
    np.random.seed(42)

    spec = Spectrum1D(spectral_axis=np.arange(10) * u.AA,
                      flux=np.random.sample(10) * u.Jy,
                      uncertainty=StdDevUncertainty(np.random.sample(10) * 0.01))

    lf = line_flux(spec)

    assert quantity_allclose(lf, u.Quantity(5.20136736, u.Jy * u.AA))
    assert quantity_allclose(lf.uncertainty, u.Quantity(0.01544415, u.Jy * u.AA))


def test_line_flux_masked():

    np.random.seed(42)

    N = 100

    wavelengths = np.linspace(0.4, 1.05, N) * u.um

    g = models.Gaussian1D(amplitude=2000*u.mJy, mean=0.56*u.um, stddev=0.01*u.um)
    flux = g(wavelengths) + 1000 * u.mJy
    noise = 400 * np.random.random(flux.shape)* u.mJy
    flux += noise

    spectrum = Spectrum1D(spectral_axis=wavelengths, flux=flux)
    spectrum.uncertainty = StdDevUncertainty(noise)

    spectrum_masked = snr_threshold(spectrum, 10.)

    # Ensure we have at least 50% of the data being masked.
    assert len(np.where(spectrum_masked.mask)[0]) > N/2

    result = line_flux(spectrum_masked)

    assert result.unit.is_equivalent(u.Jy * u.um)

    assert quantity_allclose(result.value, 720.52992, atol=0.001)

    # With flux conserving resampler
    result = line_flux(spectrum_masked, mask_interpolation=FluxConservingResampler)
    assert quantity_allclose(result.value, 720.61116, atol=0.001)


def test_line_flux_uncertainty():
    np.random.seed(42)

    spec = Spectrum1D(spectral_axis=np.arange(10) * u.AA,
                      flux=np.random.sample(10) * u.Jy,
                      uncertainty=StdDevUncertainty(np.random.sample(10) * 0.01))

    lf = line_flux(spec)

    assert quantity_allclose(lf, u.Quantity(5.20136736, u.Jy * u.AA))
    assert quantity_allclose(lf.uncertainty, u.Quantity(0.01544415, u.Jy * u.AA))


def test_equivalent_width():

    np.random.seed(42)

    frequencies = np.linspace(100, 1, 10000) * u.GHz
    g = models.Gaussian1D(amplitude=1*u.Jy, mean=10*u.GHz, stddev=1*u.GHz)
    noise = np.random.normal(0., 0.01, frequencies.shape) * u.Jy
    flux = g(frequencies) + noise + 1*u.Jy

    spectrum = Spectrum1D(spectral_axis=frequencies, flux=flux)

    result = equivalent_width(spectrum)

    assert result.unit.is_equivalent(spectrum.wcs.unit, equivalencies=u.spectral())

    # Since this is an emission line, we expect the equivalent width value to
    # be negative
    expected = -(np.sqrt(2*np.pi) * u.GHz)

    assert quantity_allclose(result, expected, atol=0.0025*u.GHz)


def test_equivalent_width_masked ():

    np.random.seed(42)

    N = 100

    wavelengths = np.linspace(0.4, 1.05, N) * u.um

    g = models.Gaussian1D(amplitude=2000*u.mJy, mean=0.56*u.um, stddev=0.01*u.um)
    flux = g(wavelengths) + 1000 * u.mJy
    noise = 400 * np.random.random(flux.shape)* u.mJy
    flux += noise

    spectrum = Spectrum1D(spectral_axis=wavelengths, flux=flux)
    spectrum.uncertainty = StdDevUncertainty(noise)

    spectrum_masked = snr_threshold(spectrum, 10.)

    # Ensure we have at least 50% of the data being masked.
    assert len(np.where(spectrum_masked.mask)[0]) > N/2

    result = equivalent_width(spectrum_masked)

    assert result.unit.is_equivalent(spectrum.wcs.unit)

    assert quantity_allclose(result.value, -719.90618, atol=0.0005)

    # With flux conserving resampler
    result = equivalent_width(spectrum_masked,
                              mask_interpolation=FluxConservingResampler)
    assert quantity_allclose(result.value, -719.89962, atol=0.0005)


def test_equivalent_width_regions():

    np.random.seed(42)

    frequencies = np.linspace(100, 1, 10000) * u.GHz
    g = models.Gaussian1D(amplitude=1*u.Jy, mean=20*u.GHz, stddev=1*u.GHz)
    noise = np.random.normal(0., 0.001, frequencies.shape) * u.Jy
    flux = g(frequencies) + noise + 1*u.Jy

    spec = Spectrum1D(spectral_axis=frequencies, flux=flux)
    result = equivalent_width(spec, regions=SpectralRegion(60*u.GHz, 10*u.GHz))

    expected = -(np.sqrt(2*np.pi) * u.GHz)

    assert quantity_allclose(result, expected, atol=0.005*u.GHz)


@pytest.mark.parametrize('continuum', [1*u.Jy, 2*u.Jy, 5*u.Jy])
def test_equivalent_width_continuum(continuum):

    np.random.seed(42)

    frequencies = np.linspace(100, 1, 10000) * u.GHz
    g = models.Gaussian1D(amplitude=1*u.Jy, mean=10*u.GHz, stddev=1*u.GHz)
    noise = np.random.normal(0., 0.01, frequencies.shape) * u.Jy
    flux = g(frequencies) + noise + continuum

    spectrum = Spectrum1D(spectral_axis=frequencies, flux=flux)

    result = equivalent_width(spectrum, continuum=continuum)

    assert result.unit.is_equivalent(spectrum.wcs.unit, equivalencies=u.spectral())

    # Since this is an emission line, we expect the equivalent width value to
    # be negative
    expected = -(np.sqrt(2*np.pi) * u.GHz) / continuum.value

    assert quantity_allclose(result, expected, atol=0.005*u.GHz)


def test_equivalent_width_absorption():

    np.random.seed(42)

    frequencies = np.linspace(100, 1, 10000) * u.GHz
    amplitude = 0.5
    g = models.Gaussian1D(amplitude=amplitude*u.Jy, mean=10*u.GHz, stddev=1*u.GHz)
    continuum = 1*u.Jy
    noise = np.random.normal(0., 0.01, frequencies.shape) * u.Jy
    flux = continuum - g(frequencies) + noise

    spectrum = Spectrum1D(spectral_axis=frequencies, flux=flux)

    result = equivalent_width(spectrum)

    assert result.unit.is_equivalent(spectrum.wcs.unit, equivalencies=u.spectral())

    expected = amplitude*np.sqrt(2*np.pi) * u.GHz

    assert quantity_allclose(result, expected, atol=0.005*u.GHz)


@pytest.mark.parametrize('bin_specification', ["centers", "edges"])
def test_equivalent_width_bin_edges(bin_specification):
    """
    Test spectrum with bin specifications as centers or edges, and at modest sampling.
    """

    np.random.seed(42)

    wavelengths = np.linspace(500, 2500, 401) * u.AA
    spectral_axis = SpectralAxis(wavelengths, bin_specification=bin_specification)

    flunit = u.Unit('W m-2 AA-1')
    amplitude = 0.5
    stddev = 100*u.AA
    expected = amplitude*np.sqrt(2*np.pi) * stddev

    g = models.Gaussian1D(amplitude=amplitude*flunit, mean=1000*u.AA, stddev=stddev)
    continuum = 1 * flunit
    noise = np.random.normal(0., 0.01, spectral_axis.shape) * flunit
    flux = continuum - g(spectral_axis) + noise

    spectrum = Spectrum1D(spectral_axis=spectral_axis, flux=flux)

    result = equivalent_width(spectrum)

    assert result.unit.is_equivalent(spectrum.wcs.unit, equivalencies=u.spectral())
    assert quantity_allclose(result, expected, atol=0.5*u.AA)

    # With SpectralRegion; need higher tolerance as this is cutting more of the wings.
    result = equivalent_width(spectrum, regions=SpectralRegion(750*u.AA, 1500*u.AA))

    assert quantity_allclose(result, expected, atol=1*u.AA)


def test_snr(simulated_spectra):
    """
    Test the simple version of the spectral SNR.
    """

    np.random.seed(42)

    #
    #  Set up the data and add the uncertainty and calculate the expected SNR
    #

    spectrum = simulated_spectra.s1_um_mJy_e1
    uncertainty = StdDevUncertainty(0.1*np.random.random(len(spectrum.flux))*u.mJy)
    spectrum.uncertainty = uncertainty

    flux = spectrum.flux

    spec_snr_expected = np.mean(flux / (uncertainty.array*uncertainty.unit))

    #
    # SNR of the whole spectrum
    #

    spec_snr = snr(spectrum)

    assert isinstance(spec_snr, u.Quantity)
    assert np.allclose(spec_snr.value, spec_snr_expected.value)


def test_snr_masked(simulated_spectra):
    """
    Test the simple version of the spectral SNR, with masked spectrum.
    """

    np.random.seed(42)

    spectrum = simulated_spectra.s1_um_mJy_e1_masked
    uncertainty = StdDevUncertainty(0.1*np.random.random(len(spectrum.flux))*u.mJy)
    spectrum.uncertainty = uncertainty

    uncertainty_array = uncertainty.array[~spectrum.mask]
    flux = spectrum.flux[~spectrum.mask]

    spec_snr_expected = np.mean(flux / (uncertainty_array * uncertainty.unit))

    spec_snr = snr(spectrum)

    assert isinstance(spec_snr, u.Quantity)
    assert np.allclose(spec_snr.value, spec_snr_expected.value)


def test_snr_no_uncertainty(simulated_spectra):
    """
    Test the simple version of the spectral SNR.
    """

    #
    #  Set up the data and add the uncertainty and calculate the expected SNR
    #

    spectrum = simulated_spectra.s1_um_mJy_e1

    with pytest.raises(Exception) as e_info:
        _ = snr(spectrum)


def test_snr_multiple_flux(simulated_spectra):
    """
    Test the simple version of the spectral SNR, with multiple flux per single dispersion.
    """

    np.random.seed(42)

    #
    #  Set up the data and add the uncertainty and calculate the expected SNR
    #

    uncertainty = StdDevUncertainty(0.1*np.random.random((5, 10))*u.mJy)
    spec = Spectrum1D(spectral_axis=np.arange(10) * u.AA,
                      flux=np.random.sample((5, 10)) * u.Jy,
                      uncertainty=uncertainty)
    snr_spec = snr(spec)
    assert np.allclose(np.array(snr_spec), [18.20863867, 31.89475309, 14.51598119, 22.24603204, 32.01461421])

    uncertainty = StdDevUncertainty(0.1*np.random.random(10)*u.mJy)
    spec = Spectrum1D(spectral_axis=np.arange(10) * u.AA, flux=np.random.sample(10) * u.Jy, uncertainty=uncertainty)
    snr_spec = snr(spec)

    assert np.allclose(np.array(snr_spec), 31.325265361800415)


def test_snr_single_region(simulated_spectra):
    """
    Test the simple version of the spectral SNR over a region of the spectrum.
    """

    np.random.seed(42)

    region = SpectralRegion(0.52*u.um, 0.59*u.um)

    #
    #  Set up the data
    #

    spectrum = simulated_spectra.s1_um_mJy_e1
    uncertainty = StdDevUncertainty(0.1*np.random.random(len(spectrum.flux))*u.mJy)
    spectrum.uncertainty = uncertainty

    wavelengths = spectrum.spectral_axis
    flux = spectrum.flux

    # +1 because we want to include it in the calculation
    l = np.nonzero(wavelengths>region.lower)[0][0]
    r = np.nonzero(wavelengths<region.upper)[0][-1]+1

    spec_snr_expected = np.mean(flux[l:r] / (uncertainty.array[l:r]*uncertainty.unit))

    #
    # SNR of the whole spectrum
    #

    spec_snr = snr(spectrum, region)

    assert np.allclose(spec_snr.value, spec_snr_expected.value)


def test_snr_two_regions(simulated_spectra):
    """
    Test the simple version of the spectral SNR within two regions.
    """

    np.random.seed(42)

    #
    # Set the regions over which the SNR is calculated
    #

    regions = [SpectralRegion(0.52*u.um, 0.59*u.um), SpectralRegion(0.8*u.um, 0.9*u.um)]

    #
    #  Set up the data
    #

    spectrum = simulated_spectra.s1_um_mJy_e1
    uncertainty = StdDevUncertainty(0.1*np.random.random(len(spectrum.flux))*u.Jy)
    spectrum.uncertainty = uncertainty

    wavelengths = spectrum.spectral_axis
    flux = spectrum.flux

    spec_snr_expected = []
    for region in regions:

        l = np.nonzero(wavelengths >= region.lower)[0][0]
        r = np.nonzero(wavelengths <= region.upper)[0][-1]+1

        spec_snr_expected.append(np.mean(flux[l:r] / (uncertainty.array[l:r]*uncertainty.unit)))

    #
    # SNR of the whole spectrum
    #

    spec_snr = snr(spectrum, regions)

    assert np.allclose(spec_snr, spec_snr_expected)


def test_snr_derived():
    np.random.seed(42)

    x = np.arange(1, 101) * u.um
    y = np.random.random(len(x))*u.Jy

    spectrum = Spectrum1D(spectral_axis=x, flux=y)

    assert np.allclose(snr_derived(spectrum), 1.604666860424951)

    sr = SpectralRegion(38*u.um, 48*u.um)
    assert np.allclose(snr_derived(spectrum, sr), 2.330463630828406)

    sr2 = SpectralRegion(48*u.um, 57*u.um)
    assert np.allclose(snr_derived(spectrum, [sr, sr2]), [2.330463630828406, 2.9673559890209305])


def test_snr_derived_masked():
    np.random.seed(42)

    x = np.arange(1, 101) * u.um
    y = np.random.random(len(x))*u.Jy
    mask = (np.random.randn(x.shape[0])) > 0
    spectrum = Spectrum1D(spectral_axis=x, flux=y, mask=mask)

    assert np.allclose(snr_derived(spectrum), 2.08175408)

    sr = SpectralRegion(38*u.um, 48*u.um)
    assert np.allclose(snr_derived(spectrum, sr), 4.01610033)

    sr2 = SpectralRegion(48*u.um, 57*u.um)
    assert np.allclose(snr_derived(spectrum, [sr, sr2]), [4.01610033, 1.94906157])


def test_centroid(simulated_spectra):
    """
    Test the simple version of the spectral centroid.
    """

    np.random.seed(42)

    #
    #  Set up the data and add the uncertainty and calculate the expected SNR
    #

    spectrum = simulated_spectra.s1_um_mJy_e1
    uncertainty = StdDevUncertainty(0.1*np.random.random(len(spectrum.flux))*u.mJy)
    spectrum.uncertainty = uncertainty

    wavelengths = spectrum.spectral_axis
    flux = spectrum.flux

    spec_centroid_expected = np.sum(flux * wavelengths) / np.sum(flux)

    #
    # SNR of the whole spectrum
    #

    spec_centroid = centroid(spectrum, None)

    assert isinstance(spec_centroid, u.Quantity)
    assert np.allclose(spec_centroid.value, spec_centroid_expected.value)


def test_centroid_masked(simulated_spectra):
    """
    Test centroid with masked spectrum.
    """

    np.random.seed(42)

    # Same as in test for unmasked spectrum, but using
    # masked version of same spectrum.
    spectrum = simulated_spectra.s1_um_mJy_e1_masked
    uncertainty = StdDevUncertainty(0.1*np.random.random(len(spectrum.flux))*u.mJy)
    spectrum.uncertainty = uncertainty

    # Use masked flux and dispersion arrays to compute
    # the expected value for centroid.
    wavelengths = spectrum.spectral_axis[~spectrum.mask]
    flux = spectrum.flux[~spectrum.mask]

    spec_centroid_expected = np.sum(flux * wavelengths) / np.sum(flux)

    spec_centroid = centroid(spectrum, None)

    assert isinstance(spec_centroid, u.Quantity)
    assert np.allclose(spec_centroid.value, spec_centroid_expected.value)


def test_inverted_centroid(simulated_spectra):
    """
    Ensures the centroid calculation also works for *inverted* spectra - i.e.
    continuum-subtracted absorption lines.
    """
    spectrum = simulated_spectra.s1_um_mJy_e1
    spec_centroid_expected = (np.sum(spectrum.flux * spectrum.spectral_axis) /
                              np.sum(spectrum.flux))

    spectrum_inverted = Spectrum1D(spectral_axis=spectrum.spectral_axis,
                                   flux=-spectrum.flux)
    spec_centroid_inverted = centroid(spectrum_inverted, None)
    assert np.allclose(spec_centroid_inverted.value, spec_centroid_expected.value)


def test_inverted_centroid_masked(simulated_spectra):
    """
    Ensures the centroid calculation also works for *inverted* spectra with
    masked data - i.e. continuum-subtracted absorption lines.
    """
    spectrum = simulated_spectra.s1_um_mJy_e1_masked
    spec_centroid_expected = (np.sum(spectrum.flux[~spectrum.mask] *
                                     spectrum.spectral_axis[~spectrum.mask]) /
                              np.sum(spectrum.flux[~spectrum.mask]))

    spectrum_inverted = Spectrum1D(spectral_axis=spectrum.spectral_axis,
                                   flux=-spectrum.flux,
                                   mask=spectrum.mask)

    spec_centroid_inverted = centroid(spectrum_inverted, None)
    assert np.allclose(spec_centroid_inverted.value, spec_centroid_expected.value)


def test_centroid_multiple_flux(simulated_spectra):
    """
    Test the simple version of the spectral SNR, with multiple flux per single dispersion.
    """

    #
    #  Set up the data and add the uncertainty and calculate the expected SNR
    #

    np.random.seed(42)

    spec = Spectrum1D(spectral_axis=np.arange(10) * u.um,
                      flux=np.random.sample((5, 10)) * u.Jy)

    centroid_spec = centroid(spec, None)

    assert np.allclose(centroid_spec.value, np.array([4.46190995, 4.17223565, 4.37778249, 4.51595259, 4.7429066]))
    assert centroid_spec.unit == u.um


def test_gaussian_sigma_width():

    np.random.seed(42)

    # Create a (centered) gaussian spectrum for testing
    mean = 5
    frequencies = np.linspace(0, mean*2, 100) * u.GHz
    g1 = models.Gaussian1D(amplitude=5*u.Jy, mean=mean*u.GHz, stddev=0.8*u.GHz)

    spectrum = Spectrum1D(spectral_axis=frequencies, flux=g1(frequencies))

    result = gaussian_sigma_width(spectrum)

    assert quantity_allclose(result, g1.stddev, atol=0.01*u.GHz)


def test_gaussian_sigma_width_masked():

    np.random.seed(42)

    # Create a (centered) gaussian masked spectrum for testing
    mean = 5
    frequencies = np.linspace(0, mean*2, 100) * u.GHz
    g1 = models.Gaussian1D(amplitude=5*u.Jy, mean=mean*u.GHz, stddev=0.8*u.GHz)
    uncertainty = StdDevUncertainty(0.1*np.random.random(len(frequencies))*u.Jy)

    mask = (np.random.randn(frequencies.shape[0]) + 1.) > 0

    spectrum = Spectrum1D(spectral_axis=frequencies, flux=g1(frequencies),
                          uncertainty=uncertainty, mask=mask)

    result = gaussian_sigma_width(spectrum)

    assert quantity_allclose(result, g1.stddev, atol=0.01*u.GHz)


def test_gaussian_sigma_width_regions():

    np.random.seed(42)

    frequencies = np.linspace(100, 0, 10000) * u.GHz
    g1 = models.Gaussian1D(amplitude=5*u.Jy, mean=10*u.GHz, stddev=0.8*u.GHz)
    g2 = models.Gaussian1D(amplitude=5*u.Jy, mean=2*u.GHz, stddev=0.3*u.GHz)
    g3 = models.Gaussian1D(amplitude=5*u.Jy, mean=70*u.GHz, stddev=10*u.GHz)

    compound = g1 + g2 + g3
    spectrum = Spectrum1D(spectral_axis=frequencies, flux=compound(frequencies))

    region1 = SpectralRegion(15*u.GHz, 5*u.GHz)
    result1 = gaussian_sigma_width(spectrum, regions=region1)

    exp1 = g1.stddev
    assert quantity_allclose(result1, exp1, atol=0.25*exp1)

    region2 = SpectralRegion(3*u.GHz, 1*u.GHz)
    result2 = gaussian_sigma_width(spectrum, regions=region2)

    exp2 = g2.stddev
    assert quantity_allclose(result2, exp2, atol=0.25*exp2)

    region3 = SpectralRegion(100*u.GHz, 40*u.GHz)
    result3 = gaussian_sigma_width(spectrum, regions=region3)

    exp3 = g3.stddev
    assert quantity_allclose(result3, exp3, atol=0.25*exp3)

    # Test using a list of regions
    result_list = gaussian_sigma_width(spectrum, regions=[region1, region2, region3])
    for model, result in zip((g1, g2, g3), result_list):
        exp = model.stddev
        assert quantity_allclose(result, exp, atol=0.25*exp)


def test_gaussian_sigma_width_multi_spectrum():

    np.random.seed(42)

    frequencies = np.linspace(100, 0, 10000) * u.GHz
    g1 = models.Gaussian1D(amplitude=5*u.Jy, mean=50*u.GHz, stddev=0.8*u.GHz)
    g2 = models.Gaussian1D(amplitude=5*u.Jy, mean=50*u.GHz, stddev=5*u.GHz)
    g3 = models.Gaussian1D(amplitude=5*u.Jy, mean=50*u.GHz, stddev=10*u.GHz)

    flux = np.ndarray((3, len(frequencies))) * u.Jy

    flux[0] = g1(frequencies)
    flux[1] = g2(frequencies)
    flux[2] = g3(frequencies)

    spectra = Spectrum1D(spectral_axis=frequencies, flux=flux)

    results = gaussian_sigma_width(spectra)

    expected = (g1.stddev, g2.stddev, g3.stddev)
    for result, exp in zip(results, expected):
        assert quantity_allclose(result, exp, atol=0.25*exp)


def test_gaussian_fwhm():

    np.random.seed(42)

    # Create a (centered) gaussian spectrum for testing
    mean = 5
    frequencies = np.linspace(0, mean*2, 100) * u.GHz
    g1 = models.Gaussian1D(amplitude=5*u.Jy, mean=mean*u.GHz, stddev=0.8*u.GHz)

    spectrum = Spectrum1D(spectral_axis=frequencies, flux=g1(frequencies))

    result = gaussian_fwhm(spectrum)

    expected = g1.stddev * gaussian_sigma_to_fwhm
    assert quantity_allclose(result, expected, atol=0.01*u.GHz)


def test_gaussian_fwhm_masked():

    np.random.seed(42)

    # Create a (centered) gaussian masked spectrum for testing
    mean = 5
    frequencies = np.linspace(0, mean*2, 100) * u.GHz
    g1 = models.Gaussian1D(amplitude=5*u.Jy, mean=mean*u.GHz, stddev=0.8*u.GHz)
    uncertainty = StdDevUncertainty(0.1*np.random.random(len(frequencies))*u.Jy)

    mask = (np.random.randn(frequencies.shape[0]) + 1.) > 0

    spectrum = Spectrum1D(spectral_axis=frequencies, flux=g1(frequencies),
                          uncertainty=uncertainty, mask=mask)

    result = gaussian_fwhm(spectrum)

    expected = g1.stddev * gaussian_sigma_to_fwhm
    assert quantity_allclose(result, expected, atol=0.01*u.GHz)


@pytest.mark.parametrize('mean', range(3,8))
def test_gaussian_fwhm_uncentered(mean):

    np.random.seed(42)

    # Create an uncentered gaussian spectrum for testing
    frequencies = np.linspace(0, 10, 1000) * u.GHz
    g1 = models.Gaussian1D(amplitude=5*u.Jy, mean=mean*u.GHz, stddev=0.8*u.GHz)

    spectrum = Spectrum1D(spectral_axis=frequencies, flux=g1(frequencies))

    result = gaussian_fwhm(spectrum)

    expected = g1.stddev * gaussian_sigma_to_fwhm
    assert quantity_allclose(result, expected, atol=0.05*u.GHz)


def test_fwhm_masked():

    np.random.seed(42)

    # Create a masked (uncentered) spectrum for testing
    frequencies = np.linspace(0, 10, 1000) * u.GHz
    stddev = 0.8*u.GHz
    g1 = models.Gaussian1D(amplitude=5*u.Jy, mean=2*u.GHz, stddev=stddev)
    mask = (np.random.randn(frequencies.shape[0]) + 1.) > 0

    spectrum = Spectrum1D(spectral_axis=frequencies, flux=g1(frequencies),
                          mask=mask)

    result = fwhm(spectrum)

    expected = stddev * gaussian_sigma_to_fwhm
    assert quantity_allclose(result, expected, atol=0.01*u.GHz)

    # Highest point at the first point
    wavelengths = np.linspace(1, 10, 100) * u.um
    flux = (1.0 / wavelengths.value) * u.Jy  # highest point first.

    spectrum = Spectrum1D(spectral_axis=wavelengths, flux=flux)
    result = fwhm(spectrum)
    # Note that this makes a little more sense than the previous version;
    # since the maximum value occurs at wavelength=1, and the half-value of
    # flux (0.5) occurs at exactly wavelength=2, the result should be
    # exactly 1 (2 - 1).
    assert result == 1.0 * u.um

    # Test the interpolation used in FWHM for wavelength values that are not
    # on the grid
    wavelengths = np.linspace(1, 10, 31) * u.um
    flux = (1.0 / wavelengths.value) * u.Jy  # highest point first.

    spectrum = Spectrum1D(spectral_axis=wavelengths, flux=flux)
    result = fwhm(spectrum)

    assert quantity_allclose(result, 1.01 * u.um)

    # Highest point at the last point
    wavelengths = np.linspace(1, 10, 100) * u.um
    flux = wavelengths.value*u.Jy # highest point last.

    spectrum = Spectrum1D(spectral_axis=wavelengths, flux=flux)
    result = fwhm(spectrum)
    assert result == 5*u.um

    # Flat spectrum
    wavelengths = np.linspace(1, 10, 100) * u.um
    flux = np.ones(wavelengths.shape)*u.Jy # highest point last.

    spectrum = Spectrum1D(spectral_axis=wavelengths, flux=flux)
    result = fwhm(spectrum)
    assert result == 9*u.um


def test_fwhm():

    np.random.seed(42)

    # Create an (uncentered) spectrum for testing
    frequencies = np.linspace(0, 10, 1000) * u.GHz
    stddev = 0.8*u.GHz
    g1 = models.Gaussian1D(amplitude=5*u.Jy, mean=2*u.GHz, stddev=stddev)

    spectrum = Spectrum1D(spectral_axis=frequencies, flux=g1(frequencies))

    result = fwhm(spectrum)

    expected = stddev * gaussian_sigma_to_fwhm
    assert quantity_allclose(result, expected, atol=0.01*u.GHz)

    # Highest point at the first point
    wavelengths = np.linspace(1, 10, 100) * u.um
    flux = (1.0 / wavelengths.value) * u.Jy  # highest point first.

    spectrum = Spectrum1D(spectral_axis=wavelengths, flux=flux)
    result = fwhm(spectrum)
    # Note that this makes a little more sense than the previous version;
    # since the maximum value occurs at wavelength=1, and the half-value of
    # flux (0.5) occurs at exactly wavelength=2, the result should be
    # exactly 1 (2 - 1).
    assert result == 1.0 * u.um

    # Test the interpolation used in FWHM for wavelength values that are not
    # on the grid
    wavelengths = np.linspace(1, 10, 31) * u.um
    flux = (1.0 / wavelengths.value) * u.Jy  # highest point first.

    spectrum = Spectrum1D(spectral_axis=wavelengths, flux=flux)
    result = fwhm(spectrum)

    assert quantity_allclose(result, 1.01 * u.um)

    # Highest point at the last point
    wavelengths = np.linspace(1, 10, 100) * u.um
    flux = wavelengths.value*u.Jy # highest point last.

    spectrum = Spectrum1D(spectral_axis=wavelengths, flux=flux)
    result = fwhm(spectrum)
    assert result == 5*u.um

    # Flat spectrum
    wavelengths = np.linspace(1, 10, 100) * u.um
    flux = np.ones(wavelengths.shape)*u.Jy # highest point last.

    spectrum = Spectrum1D(spectral_axis=wavelengths, flux=flux)
    result = fwhm(spectrum)
    assert result == 9*u.um


def test_fwhm_multi_spectrum():

    np.random.seed(42)

    frequencies = np.linspace(0, 100, 10000) * u.GHz
    stddevs = [0.8, 5, 10]*u.GHz
    g1 = models.Gaussian1D(amplitude=5*u.Jy, mean=5*u.GHz, stddev=stddevs[0])
    g2 = models.Gaussian1D(amplitude=5*u.Jy, mean=50*u.GHz, stddev=stddevs[1])
    g3 = models.Gaussian1D(amplitude=5*u.Jy, mean=83*u.GHz, stddev=stddevs[2])

    flux = np.ndarray((3, len(frequencies))) * u.Jy

    flux[0] = g1(frequencies)
    flux[1] = g2(frequencies)
    flux[2] = g3(frequencies)

    spectra = Spectrum1D(spectral_axis=frequencies, flux=flux)

    results = fwhm(spectra)

    expected = stddevs * gaussian_sigma_to_fwhm
    assert quantity_allclose(results, expected, atol=0.01*u.GHz)


def test_fwzi():
    np.random.seed(42)

    disp = np.linspace(0, 100, 1000) * u.AA

    g = models.Gaussian1D(mean=np.mean(disp),
                          amplitude=1 * u.Jy,
                          stddev=2 * u.AA)

    flux = g(disp)
    flux_with_noise = g(disp) + ((np.random.sample(disp.size) - 0.5) * 0.1) * u.Jy

    spec = Spectrum1D(spectral_axis=disp, flux=flux)
    spec_with_noise = Spectrum1D(spectral_axis=disp, flux=flux_with_noise)

    assert quantity_allclose(fwzi(spec), 226.89732509 * u.AA)
    assert quantity_allclose(fwzi(spec_with_noise), 106.99998944 * u.AA)


def test_fwzi_masked():
    np.random.seed(42)

    disp = np.linspace(0, 100, 100) * u.AA

    g = models.Gaussian1D(mean=np.mean(disp),
                          amplitude=1 * u.Jy,
                          stddev=10 * u.AA)
    flux = g(disp) + ((np.random.sample(disp.size) - 0.5) * 0.1) * u.Jy

    # Add mask. It is built such that about 50% of the data points
    # on and around the Gaussian peak are masked out (this was checked
    # with the debugger to examine in-memory data).
    uncertainty = StdDevUncertainty(0.1*np.random.random(len(disp))*u.Jy)
    mask = (np.random.randn(disp.shape[0]) - 0.5) > 0

    spec = Spectrum1D(spectral_axis=disp, flux=flux, uncertainty=uncertainty,
                      mask=mask)

    assert quantity_allclose(fwzi(spec), 35.9996284 * u.AA)


def test_fwzi_multi_spectrum():
    np.random.seed(42)

    disp = np.linspace(0, 100, 1000) * u.AA

    amplitudes = [0.1, 1, 10] * u.Jy
    means = [25, 50, 75] * u.AA
    stddevs = [1, 5, 10] * u.AA
    params = list(zip(amplitudes, means, stddevs))

    flux = np.zeros(shape=(3, len(disp)))

    for i in range(3):
        flux[i] = models.Gaussian1D(*params[i])(disp)

    spec = Spectrum1D(spectral_axis=disp, flux=flux * u.Jy)

    expected = [113.51706001 * u.AA, 567.21252727 * u.AA, 499.5024546 * u.AA]

    assert quantity_allclose(fwzi(spec), expected)


def test_is_continuum_below_threshold():

    # No mask, no uncertainty
    wavelengths = [300, 500, 1000] * u.nm
    data = [0.001, -0.003, 0.003] * u.Jy
    spectrum = Spectrum1D(spectral_axis=wavelengths, flux=data)
    assert True == is_continuum_below_threshold(spectrum, threshold=0.1*u.Jy)

#    # No mask, no uncertainty, threshold is float
#    wavelengths = [300, 500, 1000] * u.nm
#    data = [0.0081, 0.0043, 0.0072] * u.Jy
#    spectrum = Spectrum1D(spectral_axis=wavelengths, flux=data)
#    assert True == is_continuum_below_threshold(spectrum, threshold=0.1)

    # No mask, with uncertainty
    wavelengths = [300, 500, 1000] * u.nm
    data = [0.03, 0.029, 0.031] * u.Jy
    uncertainty = StdDevUncertainty([1.01, 1.03, 1.01] * u.Jy)
    spectrum = Spectrum1D(spectral_axis=wavelengths, flux=data,
                          uncertainty=uncertainty)

    assert True == is_continuum_below_threshold(spectrum, threshold=0.1*u.Jy)

    # With mask, with uncertainty
    wavelengths = [300, 500, 1000] * u.nm
    data = [0.01, 1.029, 0.013] * u.Jy
    mask = np.array([False, True, False])
    uncertainty = StdDevUncertainty([1.01, 1.13, 1.1] * u.Jy)
    spectrum = Spectrum1D(spectral_axis=wavelengths, flux=data,
                          uncertainty=uncertainty, mask=mask)

    assert True == is_continuum_below_threshold(spectrum, threshold=0.1*u.Jy)

    # With mask, with uncertainty -- should throw an exception
    wavelengths = [300, 500, 1000] * u.nm
    data = [10.03, 10.029, 10.033] * u.Jy
    mask = np.array([False, False, False])
    uncertainty = StdDevUncertainty([1.01, 1.13, 1.1] * u.Jy)
    spectrum = Spectrum1D(spectral_axis=wavelengths, flux=data,
                          uncertainty=uncertainty, mask=mask)

    print('spectrum has flux {}'.format(spectrum.flux))
    with pytest.warns(AstropyUserWarning) as e_info:
        find_lines_threshold(spectrum, noise_factor=1)
        assert len(e_info)==1 and 'if you want to suppress this warning' in e_info[0].message.args[0].lower()


def test_moment():

    np.random.seed(42)

    frequencies = np.linspace(100, 1, 10000) * u.GHz
    g = models.Gaussian1D(amplitude=1*u.Jy, mean=10*u.GHz, stddev=1*u.GHz)
    noise = np.random.normal(0., 0.01, frequencies.shape) * u.Jy
    flux = g(frequencies) + noise

    spectrum = Spectrum1D(spectral_axis=frequencies, flux=flux)

    moment_0 = moment(spectrum, order=0)
    assert moment_0.unit.is_equivalent(u.Jy )
    assert quantity_allclose(moment_0, 252.96*u.Jy, atol=0.01*u.Jy)

    moment_1 = moment(spectrum, order=1)
    assert moment_1.unit.is_equivalent(u.GHz)
    assert quantity_allclose(moment_1, 10.08*u.GHz, atol=0.01*u.GHz)

    moment_2 = moment(spectrum, order=2)
    assert moment_2.unit.is_equivalent(u.GHz**2)
    assert quantity_allclose(moment_2, 13.40*u.GHz**2, atol=0.01*u.GHz**2)

    moment_3 = moment(spectrum, order=3)
    assert moment_3.unit.is_equivalent(u.GHz**3)
    assert quantity_allclose(moment_3, 1233.78*u.GHz**3, atol=0.01*u.GHz**3)


def test_moment_cube():

    np.random.seed(42)

    frequencies = np.linspace(100, 1, 10000) * u.GHz
    g = models.Gaussian1D(amplitude=100*u.Jy, mean=50*u.GHz, stddev=1000*u.GHz)
    noise = np.random.normal(0., 0.01, frequencies.shape) * u.Jy
    flux = g(frequencies) + noise

    # use identical arrays in each spaxel. The purpose here is not to
    # check accuracy (already tested elsewhere), but dimensionality.

    flux_multid = np.broadcast_to(flux, [10,10,flux.shape[0]]) * u.Jy

    spectrum = Spectrum1D(spectral_axis=frequencies, flux=flux_multid)

    moment_1 = moment(spectrum, order=1)

    assert moment_1.shape == (10,10)
    assert moment_1.unit.is_equivalent(u.GHz)
    assert quantity_allclose(moment_1, 50.50*u.GHz, atol=0.01*u.GHz)

    # select the last axis of the cube. Should be identical with
    # the default call above.
    moment_1_last = moment(spectrum, order=1, axis=2)

    assert moment_1_last.shape == moment_1.shape
    assert moment_1_last.unit.is_equivalent(moment_1.unit)
    assert quantity_allclose(moment_1, moment_1_last, rtol=1.E-5)

    # spatial 1st order - returns the dispersion
    moment_1 = moment(spectrum, order=1, axis=1)

    assert moment_1.shape == (10,10000)
    assert moment_1.unit.is_equivalent(u.GHz)
    assert quantity_allclose(moment_1, frequencies, rtol=1.E-5)

    # higher order
    moment_2 = moment(spectrum, order=2)

    assert moment_2.shape == (10,10)
    assert moment_2.unit.is_equivalent(u.GHz**2)
    assert quantity_allclose(moment_2, 816.648*u.GHz**2, atol=0.01*u.GHz**2)

    # spatial higher order (what's the meaning of this?)
    moment_2 = moment(spectrum, order=2, axis=1)

    assert moment_2.shape == (10,10000)
    assert moment_2.unit.is_equivalent(u.GHz**2)
    # check assorted values.
    assert quantity_allclose(moment_2[0][0], 2.019e-28*u.GHz**2, rtol=0.01)
    assert quantity_allclose(moment_2[1][0], 2.019e-28*u.GHz**2, rtol=0.01)
    assert quantity_allclose(moment_2[0][3], 8.078e-28*u.GHz**2, rtol=0.01)


def test_moment_collection():

    np.random.seed(42)

    frequencies = np.linspace(100, 1, 10000) * u.GHz
    g = models.Gaussian1D(amplitude=1*u.Jy, mean=10*u.GHz, stddev=1*u.GHz)
    noise = np.random.normal(0., 0.01, frequencies.shape) * u.Jy
    flux = g(frequencies) + noise
    s1 = Spectrum1D(spectral_axis=frequencies, flux=flux)

    frequencies = np.linspace(200, 2, 10000) * u.GHz
    g = models.Gaussian1D(amplitude=2*u.Jy, mean=20*u.GHz, stddev=2*u.GHz)
    noise = np.random.normal(0., 0.02, frequencies.shape) * u.Jy
    flux = g(frequencies) + noise
    s2 = Spectrum1D(spectral_axis=frequencies, flux=flux)

    collection = SpectrumCollection.from_spectra([s1, s2])

    # Compare moments derived from collection, with moments
    # derived from individual members.
    moment_0 = moment(collection, order=0)
    moment_0_s1 = moment(s1, order=0)
    moment_0_s2 = moment(s2, order=0)
    assert quantity_allclose(moment_0[0], moment_0_s1, rtol=1.E-5)
    assert quantity_allclose(moment_0[1], moment_0_s2, rtol=1.E-5)

    moment_1 = moment(collection, order=1)
    moment_1_s1 = moment(s1, order=1)
    moment_1_s2 = moment(s2, order=1)
    assert quantity_allclose(moment_1[0], moment_1_s1, rtol=1.E-5)
    assert quantity_allclose(moment_1[1], moment_1_s2, rtol=1.E-5)

    moment_2 = moment(collection, order=2)
    moment_2_s1 = moment(s1, order=2)
    moment_2_s2 = moment(s2, order=2)
    assert quantity_allclose(moment_2[0], moment_2_s1, rtol=1.E-5)
    assert quantity_allclose(moment_2[1], moment_2_s2, rtol=1.E-5)
